local M = {}

local vector = require 'vector'
local constant = require 'constant'
local util = require 'util'

local function make_smoke_system(obj)
   local _smoke = world:atlas_entry(constant.ATLAS, 'steam')
   local _psystem = obj:add_component('CParticleSystem',
                                      {def=
                                       {n=1000,
                                        renderer={name='PSC_E2SystemRenderer',
                                                  params={entry=_smoke}},
                                        activator={name='PSConstantRateActivator',
                                                   rate=0},
                                        components={
                                           {name='PSConstantAccelerationUpdater',
                                            params={acc={0,10}}},
                                           {name='PSTimeAlphaUpdater',
                                            params={time_constant=1.5}},
                                           {name='PSBoxInitializer',
                                            params={initial={0,0,0,0},
                                                    refresh={0,0,0,0},
                                                    minv={-10,-10},
                                                    maxv={10,10}}},
                                           {name='PSConstColorInitializer',
                                            params={color=util.rgba(160, 160, 160, 255)}},
                                           {name='PSTimeInitializer',
                                            params={min_life=0.2,
                                                    max_life=3.0}},
                                           {name='PSTimeTerminator'}}}})

   local psys = _psystem:def():find_component('PSConstantRateActivator')
   local maxax = 0

   local activator = function(go, comp)
      while true do
         coroutine.yield()
         -- if the the forward direction doesn't account for most of
         -- our speed then activate the particle system
         local angle = go:angle()
         local lateral = vector.new({-math.cos(angle), -math.sin(angle)})
         local vel = vector.new(go:vel())
         local offax = vector.dot(lateral, vel)
         local thresh = 0.00001
         if offax > thresh then
            psys:rate(100 * (offax / thresh))
         else
            psys:rate(0)
         end
      end
   end
   obj:add_component('CScripted', {update_thread=util.thread(activator)})
end

function M.cons(go, m, obj, objs)
   local _truck = world:atlas_entry(constant.ATLAS, 'truck')
   local _trailer = world:atlas_entry(constant.ATLAS, 'trailer')
   local _tire = world:atlas_entry(constant.ATLAS, 'tire')
   local mw = m.width * m.tile_width
   local mh = m.height * m.tile_height

   local truck = go
   truck:add_component('CCollidable', {fixture={type='rect',
                                                w=_truck.w,
                                                h=_truck.h-8,
                                                density=10.0,
                                                offset={0,4},
                                                group=2}})

   local truck_sprite = truck:add_component('CStaticSprite', {entry=_truck})
   local angle_rate = 3.14/2
   local imp_rate = 500

   truck:fixed_rotation(0)
   truck:body_type(constant.DYNAMIC)

   local make_tire = function(truck, offset, lat)
      local tire = world:create_go()
      tire:fixed_rotation(0)
      tire:pos(vector.new(truck:pos()) + offset)
      tire:add_component('CTire', {max_lateral_impulse=lat})
      tire:add_component('CCollidable', {fixture={type='rect',
                                                  w=_tire.w,
                                                  h=_tire.h,
                                                  density=1.0,
                                                  group=2}})
      make_smoke_system(tire)
      local joint = world:create_joint(truck, offset, tire, {0,0})
      local sprite = tire:add_component('CStaticSprite',
                                        {entry=_tire,
                                         layer=constant.BACKGROUND})
      return {joint=joint, tire=tire, sprite=sprite}
   end

   local tl = make_tire(truck, {-32,12}, 100)
   local tr = make_tire(truck, {32,12}, 100)
   local bl = make_tire(truck, {-32,-12}, 3000000)
   local br = make_tire(truck, {32,-12}, 3000000)


   local trailer = world:create_go()
   local trailer_sprite = trailer:add_component('CStaticSprite', {entry=_trailer})
   trailer:pos(vector.new(truck:pos()) - {0, 128})
   trailer:add_component('CCollidable', {fixture={type='rect',
                                                  w=_trailer.w,
                                                  h=_trailer.h-8,
                                                  density=1.0,
                                                  group=2}})
   trailer:fixed_rotation(0)
   trailer:body_type(constant.DYNAMIC)
   local trl = make_tire(trailer, {-32,-10}, 3000000)
   local trr = make_tire(trailer, {32,-10}, 3000000)

   local hitch = nil
   local hitch_to = function(trailer)
      hitch = world:create_joint(truck, {0, -_truck.h/2}, trailer, {0, _trailer.h/2})
      local lim = util.d2r(90)
      hitch:limit(-lim,lim)
   end
   hitch_to(trailer)

   -- the current angle of the truck/front-tire joint and its limit
   -- (+/-)
   local angle = 0
   local limit = 1

   local controls = function(go, comp)
      while true do
         coroutine.yield()
         local dt = world:dt()
         local input = util.input_state()
         local da = -input.leftright * dt * angle_rate
         if math.abs(da) > 0 then
            angle = math.max(-limit, math.min(limit, angle + da))
         else
            angle = 0
         end

         if input.action3 and hitch then
            hitch:destroy()
            hitch = nil
         elseif not input.action3 and not hitch then
            local truck_back = util.forward_direction(truck) * (-_truck.h/2) + truck:pos()
            local trailer_front = util.forward_direction(trailer) * (_trailer.h/2) + trailer:pos()
            local dist = (trailer_front - truck_back):length()

            if dist < 24 then
               local offset = truck_back - trailer_front
               trailer:pos(offset + trailer:pos())
               hitch_to(trailer)
            end
         end

         tl.joint:limit(angle, angle)
         tr.joint:limit(angle, angle)

         tl.sprite:angle(tl.tire:angle())
         tr.sprite:angle(tr.tire:angle())
         bl.sprite:angle(bl.tire:angle())
         br.sprite:angle(br.tire:angle())
         trl.sprite:angle(trl.tire:angle())
         trr.sprite:angle(trr.tire:angle())
         truck_sprite:angle(truck:angle())
         trailer_sprite:angle(trailer:angle())

         local tangle = tl.tire:angle()
         local di = 0
         if input.action1 then
            di = 1 * dt * imp_rate
         elseif input.action2 then
            di = -1 * dt * imp_rate
         end

         local forward = util.forward_direction(tl.tire)
         local dir = forward
         local imp = dir * di
         tl.tire:apply_impulse(imp)
         tr.tire:apply_impulse(imp)

         -- update the view
         local tpos = truck:pos()
         local xmin = tpos[1] - screen_width / 2
         local xmax = tpos[1] + screen_width / 2
         local ymin = tpos[2] - screen_height / 2
         local ymax = tpos[2] + screen_height / 2
         if xmin < 0 then
            xmax = xmax + math.abs(xmin)
            xmin = 0
         elseif xmax > mw then
            xmin = xmin - (xmax - mw)
            xmax = mw
         end

         if ymin < 0 then
            ymax = ymax + math.abs(ymin)
            ymin = 0
         elseif ymax > mh then
            ymin = ymin - (ymax - mh)
            ymax = mh
         end
         set_transform(xmin, xmax, ymin, ymax)
      end
   end

   truck:add_component('CScripted', {update_thread=util.thread(controls)})
end

return M
